#include "dithering.h"

#include <stdint.h>             /* Wird für die uintX_t Datentypen benötigt */
#include <stdbool.h>             /* wird für den bool Datentyp benötigt */

/* Die 16 Dithering Sequenzen. 
 * Es werden 16 Sequenzen mit je 16 Bit angeboten
 * '0' = original DAC-Wert wird ausgegen
 * '1' = DAC-Dithering wert (DAC+1) wird ausgegeben
 * (in binärer Schreibweise - des leichteren lesens wegen ;)
 */
const uint16_t g_dither_table[16] = {
    0b0000000000000000, /* state 0 */
    0b0000000000000001, /* state 1 */
    0b0000000100000001, /* state 2 */
    0b0000010000100001, /* state 3 */
    0b0001000100010001, /* state 4 */
    0b0001001001001001, /* state 5 */
    0b0010010100100101, /* state 6 */
    0b0010101001010101, /* state 7 */
    0b0101010101010101, /* state 8 */
    0b1101101011011010, /* state 9 */
    0b1101101011011010, /* state 10 */
    0b1110110110110110, /* state 11 */
    0b1110111011101110, /* state 12 */
    0b1111101111011110, /* state 13 */
    0b1111111011111110, /* state 14 */
    0b1111111111111110  /* state 15 */
};

/* In diesem Array liegen später die 16 berechneten DAC werte für das Dithering */
uint16_t g_dac_raw_val_array[16];

/* Der Duty für das Dithering */
uint16_t g_dither_duty;

/* Prüft, ob ein spezielles Bit in der Variable gesetzt ist... 
 * IN:     [UINT16] "var" Variable, in der ein Bit geprüft werden soll 
 * IN:     [UINT8]  "bit_pos" Die Position des zu prüfenden Bits 
 * RETURN: [BOOL]   Ergebnis (true wenn '1' bzw. false wenn '0' )
 */
bool check_bit(uint16_t var, uint8_t bit_pos){
    uint16_t mask = 0u;
    uint16_t extract = 0u;

    mask = 1<<bit_pos;
    extract = (var & mask)>>bit_pos;

    if (extract == 1u){
        return true;
    }
    return false;
}

/* Berechnet welchen Wert der DAC für eine spezielle Ausgangsspannung annehemn muss
 *  IN:     [UINT32] "dac_output_in_microvolt" Die auszugebene DAC-Spannung in microvolt 1V -> 1000000uV
 *  RETURN: [UINT32] Das ganzzählige gerundete ergebnis
 */
uint32_t calulate_dac_base_steps(uint32_t dac_output_in_microvolt){
    uint32_t temp_div = 0u;

    /* Berechnun des eigentlichn DAC-Werts */
    temp_div = (dac_output_in_microvolt / DAC_1_STEP_IN_MICROVOLT);

    /* Falls das Ergebnis nicht ganz glatt ist, wird der Dithering Duty per Modulo (also aus dem Rest der Division) ermittelt */
    g_dither_duty = (((dac_output_in_microvolt % DAC_1_STEP_IN_MICROVOLT)*10000u) / DAC_1_STEP_IN_MICROVOLT);

    return temp_div;
}

/* Berechnet annäherungsweise das benötigte Dithering Muster 
 * RETURN: [UINT8] Wert entspricht das benötigte Dithering Muster aus der Dithering-Muster-Tabelle (g_dither_table) 
 */
uint8_t get_dither_sequence(void)
{
    static uint16_t sequences[16] = {
        0u,         /* seq 0 */
        625u,       /* seq 1/16 */
        1250u,      /* seq 2/16 */
        1875u,      /* seq 3/16 */
        2500u,      /* seq 4/16 */
        3125u,      /* seq 5/16 */
        3750u,      /* seq 6/16 */
        4375u,      /* seq 7/16 */
        5000u,      /* seq 8/16 */
        5625u,      /* seq 9/16 */
        6250u,      /* seq 10/16 */
        6875u,      /* seq 11/16 */
        7500u,      /* seq 12/16 */
        8125u,      /* seq 13/16 */
        8750u,      /* seq 14/16 */
        9375u,      /* seq 15/16 */
    }; 
    uint8_t loop_counter = 0u;

    for(loop_counter = 0u; loop_counter < 16u; loop_counter++){
        if(loop_counter < 15){
            if((g_dither_duty >= (sequences[loop_counter])) && (g_dither_duty < (sequences[loop_counter+1]))){
                return loop_counter;
            }
        }
    }
    return 0;
}

/* Füllt anhand des DAC-Basiswerts und der Sequenz-Zahl die Tabelle der DAC-Werte
 * IN: [UINT16] "dac_base_value" DAC Wert mit der HW-Genauigkeit
 * IN: [UINT8]  "table_sequence" Anzuwendende Dithering Sequenz
 */
void fill_dac_value_table(uint16_t dac_base_value, uint8_t table_sequence){
    uint8_t loop_counter = 0u;
    bool b_checker = false;

    for(loop_counter = 0u; loop_counter < 16u; loop_counter++){
        b_checker = check_bit(g_dither_table[table_sequence], loop_counter);
        if(b_checker == true){
            g_dac_raw_val_array[loop_counter] = (dac_base_value + 1u);
        }
        else{
            g_dac_raw_val_array[loop_counter] = dac_base_value;
        }
    }
}

uint16_t get_dithering_dac_value(uint8_t pos){
    if(pos > 15u){
        /* gibt ja nur 16 [0..15] Elemente*/
        return 0u;
    }
    return (g_dac_raw_val_array[pos]);
}

void clear_dithering_table(void){
    uint8_t loop_counter = 0u;
    for(loop_counter = 0; loop_counter < 16u; loop_counter++){
        g_dac_raw_val_array[loop_counter] = 0u;
    }
}